/*----------------
Steve Waterhouse
chol.c
Tue 31 Oct 1995
---------------*/

#include	"utils.h"
#include        "matrix.h"
#include        "chol.h"
#include	<math.h>


void cholesky_solve(real **A, int dim, real *b, real *y) 
{
    long i, j;
    real pos_def;
    real **decomp;
    real sum;
    real det;
	
    decomp = RealMatrix2(dim, dim);

/*	print_matrix(A, dim, dim);*/
	
	pos_def = cholesky_decomp(A, decomp, dim);

	while(pos_def == 0)
	{
		error(1,"not positive definite");
		for(i = 0; i < dim; i ++) A[i][i] += 1.0;
		pos_def = cholesky_decomp(A, decomp, dim);
	}
	
 	det = 2.0 * log_trace(decomp, dim);	

    for(i = 0; i < dim; i ++) /* solve L.x = b where x = L'.y
				 and store solution x in y */
	{
		sum = b[i];
		for(j = 0; j < i; j ++)
		{
			sum -= decomp[i][j] * y[j];
		}
		y[i] = sum / decomp[i][i]; /* put the solution into y */
	}
	for(i = dim - 1; i >= 0; i --) /* solve L'y = x */
	{
		sum = y[i];
		for(j = i + 1; j < dim; j++)
		{
			sum -= decomp[j][i] * y[j];
		}
		y[i] = sum / decomp[i][i];
	}
    FreeRealMatrix2(&decomp);
}

int cholesky_decomp(real **mat, real **decomp, int dim) 
{
	 /* Cholesky decomposition, returns lower triangle L, and whether 
	    matrix is positive definite or not. 
	    input  ->    mat : square matrix, dimension dim;
	    output -> decomp : lower triangle of Cholesky decomposition 
	    */
     
    long i, j, k;
    real sum;

	copy_matrix(mat, decomp, dim, dim);
	
    for(i = 0; i < dim; i ++) 
    {

        for(j = 0; j < dim; j ++)
        {
            sum  = mat[i][j];
			
            for(k = i - 1; k >= 0; k --) 
            {
                sum -= decomp[i][k] * decomp[j][k];
            }
            if(i == j) 
            {
                if( sum <= 0.0) return(0); /* not positive definite matrix */
                decomp[i][i] = sqrt( sum );
            }
            else
            {
                decomp[j][i] = sum / decomp[i][i];
            }
        }
    }
	return(1);
	
}        

void solve_ls(real *grad,real **hess,real *theta,int ipdim) {
	long singular;
	
	if(issingular(hess,ipdim)) {
		print_matrix(hess,ipdim,ipdim);
		error(1,"Singular matrix in solve_ls");
		singular = 1;  
	}
	cholesky_solve(hess,ipdim,grad,theta);
}

#define TINY2 1.0e-10


void newton(real *grad,real **hess,real *theta,int ipdim,float lambda) {
     int bad_flag;
     real *new_theta;
     int singular;
     long d;
     
     new_theta = RealVector(ipdim);
     
     if(issingular(hess,ipdim)) 
       {
	    print_matrix(hess,ipdim,ipdim);
	    error(1,"Singular matrix in newton");
	    singular = 1;  
       }
     
     copy_vector(theta,new_theta,ipdim);
     
     cholesky_solve(hess,ipdim,grad,theta);
     
     for(d=0;d<ipdim;d++)
       {
	    theta[d] *= lambda;
	    new_theta[d] += theta[d];
       }
     
     
     bad_flag = 0;
     if(!bad_flag)
       for(d=0;d<ipdim;d++)
	 if(isnan(new_theta[d]))
	   {
		error(1,"theta is nan");
		bad_flag = 1;
	   }
     if(!bad_flag) 
       {
	    for(d=0;d<ipdim;d++)
	      theta[d] = new_theta[d];
       }
     FreeRealVector(&new_theta);
}
 
