#include "incs.h"
#include "net_class_ens.h"
#include "utils.h"
#include "functions.h"
#include "lu.h"
#include "matrix.h"

NET_Class_Ens::NET_Class_Ens(int n_batch, 
		     int num_ip, 
		     int num_op,
		     int dbg,
		     int _msteps,
		     real _tol)
  :    n_weights(num_ip * num_op),
       msteps(_msteps),
       tol(_tol),
       Net(n_batch, num_ip, num_op, dbg)
{
  weights = RealVector(n_weights);
  dweights = RealVector(n_weights);
  last_weights = RealVector(n_weights);
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = 0.0;
	dweights[w] = 0.0;
	last_weights[w] = 0.0;
  }
  eta = RealVector(n_output);
  hessian = RealMatrix3(n_output, n_input,n_input);
  inv_hess = RealMatrix3(n_output, n_input,n_input);
  alpha = RealVector(n_output);
  for(int o = 0; o < n_output; o ++)
	alpha[o] = 0.0001;
  ssw = RealVector(n_output);
  hess_trace = RealVector(n_output);
  hess_det = RealVector(n_output);
  strcpy(class_str, "NET_Class_Ens");
}
  
NET_Class_Ens::~NET_Class_Ens()
{
  FreeRealVector(&weights);
  FreeRealVector(&dweights);
  FreeRealVector(&last_weights);
}

NET_Class_Ens::NET_Class_Ens(const NET_Class_Ens & old)
  :    n_weights(old.n_weights),
       msteps(old.msteps),
       tol(old.tol),
       Net(old)
{
  weights = RealVector(n_weights);
  dweights = RealVector(n_weights);
  last_weights = RealVector(n_weights);
  eta = RealVector(n_output);
  for(int i =0; i < n_weights; i ++) {
	weights[i] = old.weights[i];
  }
  hessian = RealMatrix3(n_output, n_input,n_input);
  inv_hess = RealMatrix3(n_output, n_input,n_input);
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_input; i ++) {
	  for(int i2 = 0; i2 < n_input; i2 ++) {
		hessian[o][i][i2] = old.hessian[o][i][i2];
		inv_hess[o][i][i2] = old.inv_hess[o][i][i2];
	  }
	}
  }
  alpha = RealVector(n_output);
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = old.alpha[o];
  }
  ssw = RealVector(n_output);
  for(int o = 0 ; o < n_output; o ++) {
	ssw[o] = old.ssw[o];
  }
  hess_trace = RealVector(n_output);
  for(int o = 0 ; o < n_output; o ++) {
	hess_trace[o] = old.hess_trace[o];
  }
  hess_det = RealVector(n_output);
  for(int o = 0 ; o < n_output; o ++) {
	hess_det[o] = old.hess_det[o];
  }
  strcpy(class_str, "NET_Class_Ens");
}

NET_Class_Ens::NET_Class_Ens(int n_batch, FILE* fp) 
  : Net(n_batch, fp)
{
  n_weights = n_input * n_output;
  strcpy(class_str, "NET_Class_Ens");
  weights = RealVector(n_weights);
  dweights = RealVector(n_output * n_input);
  last_weights = RealVector(n_output * n_input);
  eta = RealVector(n_output);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = (real) tmp;
    }
  }  
  hessian = RealMatrix3(n_output, n_input,n_input);
  inv_hess = RealMatrix3(n_output, n_input,n_input);
  alpha = RealVector(n_output);
  ssw = RealVector(n_output);
  hess_trace = RealVector(n_output);
  hess_det = RealVector(n_output);
}  

void NET_Class_Ens::randomize_weights(real offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
	double norm = cos(2*PI * drand48()) * sqrt(-2.0 * log(drand48()));
    weights[wt_num] = offset * norm / sqrt(n_input);
  }
}


void NET_Class_Ens::randomize_weights2(real offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
    weights[wt_num] = (1.0 + offset * 2.0*(get_drand() - 0.5)) * weights[wt_num]  ;

  }
}


void NET_Class_Ens::go_forward(real* in, real* out)
{
  go_forward_log(in, out);
  for(int op = 0;  op < n_output;  op ++) {
	out[op] = Sexp(out[op]);
	if(out[op] < VSMALL) out[op] = VSMALL;
  }
}

void NET_Class_Ens::go_forward_log(real* in, real* out)
{
  for(int o = 0;  o < n_output;  o ++) {
    eta[o] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      eta[op] += weights[op*n_input + ip] * in[ip];
    }
  }
  log_softmax(out, eta, n_output);
  if(debug > 2)
	for(int op2 = 0;  op2 < n_output;  op2 ++) 
	  check_arg(out[op2], "NET_Class_Ens: log_out");
}
void NET_Class_Ens::go_forward(real* in, real* out, real* m2)
{
  for(int op = 0;  op < n_output;  op ++) {
	m2[op] = 0.0;
  }
  go_forward(in, out);
}

double NET_Class_Ens::train(int n_points, real** in, real** targ)
{
  for(int n = 0; n < n_points; n ++) {
	jpost[n] = 1.0;
  }
  inputs = in;
  targets = targ;
  double cost = train_params(n_points);
  return cost;
}

double NET_Class_Ens::train(int n_points, real** in, real** targ, real* wts)
{
  
 for(int n = 0; n < n_points; n ++) {
   jpost[n] = Sexp(wts[n]);
   if(jpost[n] < VSMALL) jpost[n] = VSMALL;
 }
  inputs = in;
  targets = targ;
 
  for(int w1 = 0; w1 < n_weights; w1 ++) 
	last_weights[w1] = weights[w1];

  double cost =  train_params(n_points, jpost);
  for(int w = 0; w < n_weights; w ++)
	dweights[w] = weights[w] - last_weights[w];
  return cost;
}

void NET_Class_Ens::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d %d\n", n_output, n_input);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      fprintf(fp, "%f ", (float) weights[i*n_input +j]);
    }
    fprintf(fp, "\n");
  }
}

void NET_Class_Ens::read_weights(FILE* fp)
{
  int ninp;
  int noutp;
  fscanf(fp, "%s\n", class_str);
  if(strstr(class_str, "NET_Class_Ens") == NULL)
	error(1, "No class id %s found", class_str);
  fscanf(fp, "%d %d\n", &noutp, &ninp);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = (real) tmp;
    }
  }
}
void NET_Class_Ens::get_gradient(double* grad) {
  for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      int n = o * n_input + i;
      grad[n] = alpha[o] * weights[n];
    }
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
	real wt = jpost[point];
	if(wt > MIN_PROB) {
	  int wt_num = 0;
	  for(int op = 0; op < n_output; op ++) {
		double err = (tmp_op[op] - targets[point][op]);
		double wt_err = wt * err;
		for(int ip =0; ip < n_input; ip ++) {
		  grad[wt_num] +=  wt_err * inputs[point][ip];
		  wt_num ++;
		}
	  }
    }
  }
}
void NET_Class_Ens::get_gradient(double* grad, double* fn) {
  for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      int n = o * n_input + i;
      grad[n] = alpha[o] * weights[n];
    }
  }
  *fn = 0.0;
  get_new_ssw();
  for(int o = 0; o < n_output; o ++) {
	(*fn) += alpha[o]*ssw[o];
  }

  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
    int wt_num = 0;
	real wt = jpost[point];
	if(wt > MIN_PROB) {
	  for(int op = 0; op < n_output; op ++) {
		double err = (tmp_op[op] - targets[point][op]);
		double err2 = err * wt;
		for(int ip =0; ip < n_input; ip ++) {
		  grad[wt_num] +=  err2 *  inputs[point][ip];
		  wt_num ++;
		}
		(*fn) -= wt * targets[point][op] * Slog(tmp_op[op]);
	  }
	}
  }
}

double NET_Class_Ens::get_function() {
  double fn = 0.0;
  get_new_ssw();
  for(int o = 0; o < n_output; o ++) {
	(fn) += alpha[o]*ssw[o];
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], tmp_op);
	double wt = jpost[point];
	if(wt > MIN_PROB) {
	  for(int op = 0; op < n_output; op ++) {
		fn -= wt * targets[point][op] * Slog(tmp_op[op]);
	  }
	}
  }
  return fn;
}
void NET_Class_Ens::compute_hessian(int n_points, 
				  real** in, 
				  real** out, 
				  real* wts,
				  real*** hess)
{
  for(int o2 = 0; o2 < n_output; o2 ++) {
    for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
		hess[o2][i][i2] = 0.0;
      }
    }
	for(int i = 0; i < n_input; i ++) {
      hess[o2][i][i] = alpha[o2];
    }
  }
  for(int n = 0; n < n_points; n ++) {
    for(int o = 0; o < n_output; o ++) {
      real w = wts[n] * out[n][o] * (1.0 - out[n][o]);
      for(int i = 0; i < n_input; i ++) {
		for(int i2 = 0; i2 < n_input; i2 ++) {
		  hess[o][i][i2] += w  * in[n][i] * in[n][i2];
		}
      }
    }
  }
}

double NET_Class_Ens::get_joint(real* in, real* targ)
{
  double joint = 0.0;
  go_forward(in, tmp_op);
  for(int op = 0; op < n_output; op ++) {
	joint += targ[op] * Slog(tmp_op[op]);
  }
  return (joint);
}

real NET_Class_Ens::get_prior() {
  real arg = 0.0;
  for(int op = 0; op < n_output; op ++) {
	arg += 0.5*n_output * Slog(alpha[op]);
	arg -= 0.5*alpha[op] * ssw[op];
  }
  real prior = -(arg);

  return prior;
}

void NET_Class_Ens::get_new_ssw() {
  for(int o = 0; o < n_output; o ++) {
	 ssw[o] = 0.0;
	 for(int i = 0; i < n_input; i ++) {
	   ssw[o] += weights[o*n_input + i] * weights[o*n_input + i];
	 }
  }
}

void NET_Class_Ens::update_hyperparameters(int n_points, real** in, 
									   real** targ, real* wts) 
{
  go_forward_batch(n_points, in, tmp_op_vec);
  for(int o = 0; o < n_points; o ++) {
	jpost[o] = Sexp(wts[o]);
  }
  compute_hessian(n_points, in, tmp_op_vec, jpost, hessian);
  
  for(int o = 0; o < n_output; o ++){
    hess_det[o] = lu_inverse(hessian[o], inv_hess[o], n_input);
	if(isposdef(hessian[o], n_input) == 0) {
	  printf("warning...hessian is not positive definite in NET_Regress_Ens\n");
	  hess_trace[o] = 0.0;
	}
	else { 
	  hess_trace[o] = 0.0;
	  for(int i = 0; i < n_input; i ++) {
		hess_trace[o] += inv_hess[o][i][i];
	  }
	}
  }
  get_new_ssw();
  for(int o = 0; o < n_output; o ++) {
	double alpha_0 = 0.1;
	double ssw_0 = 0.1;
	alpha[o] = (n_input + alpha_0) / (ssw[o] + ssw_0 + hess_trace[o]);
	//	 printf("gate alpha = %f\n", alpha[o]);
  }
}


void NET_Class_Ens::init(int n_points, real** in, 
					 real** targ, real* wts) {

  train(n_points, in, targ, wts);
}

real NET_Class_Ens::get_sum_jposts(int n_points, real* wts) {
  real sum = LZERO;
  for (int n = 0; n < n_points; n ++) {
	sum = LAdd(sum, wts[n]);
  }
  return sum;
}

real NET_Class_Ens::get_evidence(int n_points, real** in, real** targ,
								 real* wts) {
  real evidence = 0.0;
  
  evidence += get_prior();

  for(int o = 0; o < n_output; o ++) {
	evidence -= hess_det[o] / 2.0 ;
  }
  for(int n = 0; n < n_points; n ++) 
	evidence += Sexp(wts[n]) * get_joint(in[n], targ[n]);
  
  return evidence;
}

