#include "tree.h"

Tree::~Tree()
{
  FreeRealMatrix2(&outputs);
  FreeRealMatrix2(&tmp_ops);
  FreeRealMatrix2(&tmp_mom);
  FreeRealVector(&tmp_op);
  FreeCharVector(&class_str);
  delete_nodes();
}

void Tree::delete_nodes() {
  for(int t = 0; t < n_terminals; t ++) {
	delete terminals[t];
  }
  delete [] terminals;
    
  for(int t1 = 0; t1 < n_terminals; t1 ++) {
	delete term_nodes[t1];
  }
  delete [] term_nodes;
  
  for(int nt = 0; nt < n_gates; nt ++) {
	delete gates[nt];
  }
  delete [] gates;
  
  for(int nt1 = 0; nt1 < n_gates; nt1 ++) {
	delete nt_nodes[nt1];
  }
  delete [] nt_nodes;
}
Tree::Tree(int n_batch, 
	   int num_expert, 
	   int num_depth,
	   int num_ip, 
	   int num_op,
	   int dbg,
	   real _tol,
	   int _itmax,
	   int _msteps,
	   int _rich,
	   int _seed,
	   int _gate_ensemble,
	   int _xprt_ensemble,
	   int _gate_hyper_epochs,
	   int _xprt_hyper_epochs
	   )
:	size_batch(n_batch),
    n_expert(num_expert),
    n_input(num_ip+1),
    n_output(num_op), 
    n_levels(num_depth),
    debug(dbg),
    optim_tolerance(_tol),
    optim_iterations(_itmax),
    msteps(_msteps),
    optim_richness(_rich),
    seed(_seed),
	gate_ensemble(_gate_ensemble),
	xprt_ensemble(_xprt_ensemble),
	gate_hyper_epochs(_gate_hyper_epochs),
	xprt_hyper_epochs(_xprt_hyper_epochs)

{
  class_str = CharVector(1024);

  n_nodes = 0;
  outputs = RealMatrix2(size_batch, n_output);
  tmp_ops = RealMatrix2(size_batch, n_output);
  tmp_mom = RealMatrix2(size_batch, n_output);
  tmp_op = RealVector(n_output);
  beta = 1.0;
  mac_pass = 0;
  srt_in = RealMatrix2(size_batch, n_input);
  jpost = RealMatrix2(2, size_batch);
  cpost = RealMatrix2(size_batch, 2);
  tmp = RealVector(size_batch + 1);
  indx = IntVector(size_batch + 1 );
  in_batch = RealMatrix2(size_batch, n_input);
}

Tree::Tree(int n_batch, FILE* fp, int dbg)
{
  debug = dbg;
  class_str = CharVector(1024);
  size_batch = n_batch;

  fscanf(fp, "%s\n", class_str);
  fscanf(fp, "%d %d\n", &n_input, &n_output);
  outputs = RealMatrix2(size_batch, n_output);
  tmp_ops = RealMatrix2(size_batch, n_output);
  tmp_mom = RealMatrix2(size_batch, n_output);
  tmp_op = RealVector(n_output);

  char* child_class_str = CharVector(1024);
  fscanf(fp, "%s\n", child_class_str);
  char* ref1 = "Node_NT";
  char* ref2 = "Node_Term";
  if(strcmp(child_class_str, ref1) == 0)
    node = new Node_NT(size_batch, fp);
  else if(strcmp(child_class_str, ref2) == 0)
    node = new Node_Term(size_batch, fp);
  else{
    printf("%s : %s Expected one of Node_NT or Node_Term\n", 
		   class_str, child_class_str);
    exit(0);
  }
  if(node == NULL) {
	error(1, "node is null\n");
  }
  gate_ensemble = 0;
  xprt_ensemble = 0;
  gate_hyper_epochs = 0;
  xprt_hyper_epochs = 0;
  beta = 1.0;
  FreeCharVector(&child_class_str);
  srt_in = RealMatrix2(size_batch, n_input);
  jpost = RealMatrix2(2, size_batch);
  cpost = RealMatrix2(size_batch, 2);
  tmp = RealVector(size_batch + 1);
  indx = IntVector(size_batch + 1 );
  in_batch = RealMatrix2(size_batch, n_input);

}

Tree::Tree(Tree* old) :
  size_batch(old->size_batch),
  n_expert(old->n_expert),
  n_input(old->n_input),
  n_output(old->n_output),
  n_levels(old->n_levels),

  debug(old->debug),
  optim_tolerance(old->optim_tolerance),
  optim_iterations(old->optim_iterations),
  msteps(old->msteps),
  optim_richness(old->optim_richness),
  seed(old->seed),
  n_nodes(old->n_nodes)
{
  class_str = CharVector(1024);
  strcpy(class_str,"Tree");
  outputs = RealMatrix2(size_batch, n_output);
  tmp_ops = RealMatrix2(size_batch, n_output);
  tmp_mom = RealMatrix2(size_batch, n_output);
  tmp_op = RealVector(n_output);
  beta = 1.0;
}
Tree::Tree(const Tree& old)
:	size_batch(old.size_batch),
    n_expert(old.n_expert),
    n_input(old.n_input),
    n_output(old.n_output),
    n_levels(old.n_levels),
    debug(old.debug),
    optim_tolerance(old.optim_tolerance),
    optim_iterations(old.optim_iterations),
    msteps(old.msteps),
    optim_richness(old.optim_richness),
    seed(old.seed),
	n_nodes(old.n_nodes)
{
  class_str = CharVector(1024);
  strcpy(class_str,"Tree");
  outputs = RealMatrix2(size_batch, n_output);
  tmp_ops = RealMatrix2(size_batch, n_output);
  tmp_mom = RealMatrix2(size_batch, n_output);
  tmp_op = RealVector(n_output);
  beta = 1.0;
  srt_in = RealMatrix2(size_batch, n_input);
  jpost = RealMatrix2(2, size_batch);
  cpost = RealMatrix2(size_batch, 2);
  tmp = RealVector(size_batch + 1);
  indx = IntVector(size_batch + 1 );
  in_batch = RealMatrix2(size_batch, n_input);

}

int Tree::get_new_id() {
  n_nodes ++;
  return n_nodes;
}

void Tree::count_nodes(Node* nd, int* n_terms, int* n_nonterms) {
  if(nd == NULL)
	error(1, "node is null in count_nodes");
  if(strcmp(nd->get_class(), "Node_NT") ==0) {
	(*n_nonterms) ++;
	Node_NT* nt = (Node_NT*) nd;
	int nch = nt->get_num_children();
	for(int ch = 0; ch < nch; ch ++) {
	  count_nodes(nt->get_child(ch), n_terms, n_nonterms);
	}
  }
  else {
	if(strcmp(nd->get_class(), "Node_Term") ==0) {	
	  (*n_terms) ++;
	}
	else {
	  error(1, "no such node found in count_nodes: %s %d", nd->get_class(), nd->get_id());
	}
  }
}
void Tree::count_params(Node* nd, int* n_params)
{
  if(nd == NULL)
	error(1, "node is null in count_nodes");
  if(strcmp(nd->get_class(), "Node_NT") ==0) {
	Node_NT* nt = (Node_NT*) nd;
	int nch = nt->get_num_children();
	(*n_params) += nch * n_input;
	for(int ch = 0; ch < nch; ch ++) {
	  count_params(nt->get_child(ch), n_params);
	}
  }
  else {
	if(strcmp(nd->get_class(), "Node_Term") ==0) {	
	  (*n_params) += n_output * n_input;
	}
	else {
	  error(1, "no such node found in count_nodes");
	}
  }
}
int Tree::get_n_params() {
  int n_params = 0;
  
  count_params(node, &n_params);
  macopt_init(n_params, 0);
  n_weights = n_params;
  return n_params;
}

void Tree::collect_nodes(Node* nd) {
  if(nd == NULL)
	error(1, "node is null in collect_nodes");
  if(strcmp(nd->get_class(), "Node_NT") ==0) {
	if(gate_count < n_gates) {
	  nt_nodes[gate_count] = nd;
	  Node_NT* nt = (Node_NT*) nd;
	  gates[gate_count] = nt->get_gate();
	  gate_count ++;
	}
	else{
	  error(1,"too many gates: %d %d", gate_count, n_gates);
	}
	Node_NT* nt = (Node_NT*) nd;
	int nch = nt->get_num_children();
	for(int ch = 0; ch < nch; ch ++) {
	  collect_nodes(nt->get_child(ch));
	}
  }
  else {
	if(strcmp(nd->get_class(), "Node_Term") ==0) {	
	  if(term_count < n_terminals) {
		term_nodes[term_count] = nd;
		Node_Term* nt = (Node_Term*) nd;
		terminals[term_count] = nt->get_expert();
		term_count ++;
	  }
	  else {
		error(1, "too many terminals");
	  }
	}
	else {
	  error(1, "no such node found in count_nodes");
	}
  }
}
void Tree::gather_nodes() {
  term_count = 0;
  gate_count = 0;
  n_terminals = 0;
  n_gates = 0;
  count_nodes(node, &n_terminals, &n_gates);
  //  printf("n_terminals = %d\n", n_terminals);
  //  printf("n_gates = %d\n", n_gates);
  //  printf("n_input = %d\n", n_input);
  //  printf("n_output = %d\n", n_output);
  terminals = new Network*[n_terminals];
  term_nodes = new Node*[n_terminals];
  nt_nodes = new Node*[n_gates];
  gates = new Network*[n_gates];
  for(int e = 0; e < n_terminals; e++) {
	terminals[e] = NULL;
	term_nodes[e] = NULL;
  }
  for(int g = 0; g < n_gates; g ++) {
	gates[g] = NULL;
	nt_nodes[g] = NULL;
  }
  collect_nodes(node);
}

int Tree::num_input()
{
  return n_input;
}

int Tree::num_output()
{
  return n_output;
}

Node* Tree::build_recursive_tree(Node_NT* old, Node* parent)
{

  Node_NT* this_node;
  this_node = new Node_NT(*old);
  
  // First Build the Gate 
  Network* gate;
  Network* old_gate = old->get_gate();
  gate = get_new_net(old_gate);

  this_node->set_gate(gate);

  int num_children = old->get_num_children();
  Node** children = new Node* [num_children];


  for(int xp = 0; xp < num_children; xp ++){
	Node* old_child = old->get_child(xp);
	if(strcmp(old_child->get_class(), "Node_Term") == 0) {
      Network* expert;
	  Node_Term* old_term = (Node_Term*) old_child;
      children[xp] = (Node_Term*) new Node_Term(*old_term);
	  children[xp]->set_id(get_new_id());
	  Network* xprt = ((Node_Term*) old_child)->get_expert();

	  expert = get_new_net(xprt);

	  Node_Term* this_child = (Node_Term*) children[xp];
	  this_child->set_expert(expert);
	  this_node->set_child(xp, children[xp]);
    }
    else // we have a Node_NT
      {
		children[xp] = build_recursive_tree((Node_NT*)old_child, this_node);
		this_node->set_child(xp, children[xp]);
	  }
  }

  for(int xp1 = 0 ; xp1 < num_children; xp1 ++) {
    children[xp1]->set_parent(this_node);
  }
  this_node->set_parent(parent);
  return this_node;
}

void Tree::go_forward_batch(int n_points, real** in, real** out)
{
  for(int n = 0; n < n_points; n ++) {
	node->set_joint_prob(0.0, in[n], n);
	for(int o = 0; o < n_output; o ++) {
		out[n][o] = 0.0;
	}
	for(int ch = 0 ; ch < n_terminals; ch ++) {
	  real jp = term_nodes[ch]->get_joint_prob(n);
	  terminals[ch]->go_forward(in[n], tmp_ops[n]);
	  for(int o2 = 0; o2 < n_output; o2 ++) {
		out[n][o2] += Sexp(jp) * tmp_ops[n][o2];
	  }
	}
  }
}

void Tree::go_forward_batch(int n_points, real** in, real** out,
							real** m2)
{
  for(int n = 0; n < n_points; n ++) {
	node->set_joint_prob(0.0, in[n], n);
	for(int o = 0; o < n_output; o ++) {
	  out[n][o] = 0.0;
	  m2[n][o] = 0.0;
	}
	for(int ch = 0 ; ch < n_terminals; ch ++) {
	  real jp = Sexp(term_nodes[ch]->get_joint_prob(n));
	  terminals[ch]->go_forward(in[n], tmp_ops[n], tmp_mom[n]);
	  for(int o2 = 0; o2 < n_output; o2 ++) {
		out[n][o2] += jp * tmp_ops[n][o2];
		m2[n][o2] += jp * tmp_mom[n][o2];
	  }
	}
  }
  for(int n2 = 0; n2 < n_points; n2 ++) {
	for(int op = 0; op < n_output; op ++) {
	  m2[n2][op] -= out[n2][op]*out[n2][op];
	}
  }
}
